package confutil;

import java.io.*;
import java.net.*;

/**
 * This class encapsulates the main configuration client functionality. It offers methods
 * to connect to a Linksys box where a configuration server runs, authenticate a user, get, set and unset
 * configuration variables that reside in the NVRAM of the box and control the function of the
 * P2PWNC protocol modules that run on the Linksys WRT54GS wireless router. The configuration client
 * and server programs communicate by exchanging simpe text messages over TCP.
 *
 * <p>Title: P2PWNC configuration utility</p>
 * <p>Description: Linksys WRT54GS P2PWNC configuration utility</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: </p>
 * @author pfrag
 * @version 1.0
 */
class ConfClient {
  /**
   * "GET" message. Gets the value of a specified NVRAM variable.
   */
  public static final int ACTION_GET = 10;

  /**
   * "SET" message. Gets the value of a specified NVRAM variable to a given value.
   */
  public static final int ACTION_SET = 20;

  /**
   * "UST" message. Unsets an NVRAM variable.
   */
  public static final int ACTION_UNSET = 30;

  /**
   * "ACK" message. The requested operation has been performed successfully.
   */
  public static final int ACTION_ACK = 40;

  /**
   * "ERR" message. An error occured while performing a requested operation.
   */
  public static final int ACTION_ERR = 50;

  /**
   * "AUT" message. Password based authentication to a Linksys router.
   */
  public static final int ACTION_AUTH = 60;

  /**
   * "AOK" message. User successfully authenticated.
   */
  public static final int ACTION_AUTH_OK = 70;

  /**
   * "AER" message. Authentication failed.
   */
  public static final int ACTION_AUTH_ERR = 80;

  /**
   * Default port where a configuration server listens for client messages.
   */
  public static final int CONF_SERVER_PORT = 5555;

  /**
   * Configuration server IP address.
   */
  protected InetAddress serverAddr;

  /**
   * Constructs a ConfClient object using addr as the configuration server address.
   * @param addr InetAddress
   * Configuration server address. This is the IP address of the Linksys WRT54GS router where
   * the server runs.
   */
  ConfClient(InetAddress addr) {
    this.serverAddr = addr;
  }

  /**
   * Constructs a text message to be sent to the server. Action must be one of the valid
   * actions specified. Valid messages to be sent are "GET", "SET", "UST", "AUT".
   * @param action int
   * The message type (ACTION_GET, ACTION_UNSET, ACTION_SET, ACTION_AUTH
   * @param var String
   * The "VAR" message field. In case of a "GET", "SET" or "UST" message it specifies
   * the variable whose value is to be retrieved, set or unset. In case of an "AUT" message,
   * var specifies the Linksys router's administrator password.
   * @param val String
   * The "VAL" message field. It is used only in "SET" messages, where it represents the value of the
   * variable to be set.
   * @return String
   * The constructed text message.
   */
  public String makeMessage(int action, String var, String val) {
    StringBuffer text = new StringBuffer();
    int len = 0;
    int clen = 0;

    len = 3 + "\r\nContent-length: ".length();
    switch (action) {
      case ACTION_GET:
        clen += "VAR: ".length() + var.length();
        len += Integer.toString(clen).length() + clen + 3;
        text.append("GET\r\nContent-length: " + Integer.toString(clen) + "\r\nVAR: " + var);
        return text.toString();

      case ACTION_SET:
        clen += "VAR: ".length() + var.length() + 2 + "VAL: ".length() + val.length();
        len += Integer.toString(clen).length() + clen + 3;
        text.append("SET\r\nContent-length: " + Integer.toString(clen) + "\r\nVAR: " + var + "\r\nVAL: " + val);
        return text.toString();

      case ACTION_UNSET:
        clen += "VAR: ".length() + var.length();
        len += Integer.toString(clen).length() + clen + 3;
        text.append("UST\r\nContent-length: " + Integer.toString(clen) + "\r\nVAR: " + var);
        return text.toString();

      case ACTION_AUTH:
        clen += "VAR: ".length() + var.length();
        len += Integer.toString(clen).length() + clen + 3;
        text.append("AUT\r\nContent-length: " + Integer.toString(clen) + "\r\nVAR: " + var);
        return text.toString();

      default:
        return null;
    }
  }

  /**
   * Parses a text message received by the server. Expected messages are "ACK" and "ERR". Other
   * messages that a client can receive are "AOK" and "AER", but these need not be parsed because the consist
   * only of the message type identifier ("AOK", "AER").
   * @param text String
   * The text message to be parsed.
   * @return String
   * The value of the "VAR" field. It indicates which variable was requested to be set/unset.
   */
  public String parseMessage(String text) {
    int pos = 0;
    int crlfpos = 0;
    int clen = 0;
    String var = null;

    //check message type (must be ACK/ERR)
    if (
      (text.toLowerCase().startsWith("ack\r\ncontent-length: ") == false ) &&
      (text.toLowerCase().startsWith("err\r\ncontent-length: ") == false ) ){
            System.err.println("parseMessage: Error parsing message type");
            return null;
    }
    //move position to the start of the next message header ("Content-length")
    pos = 3 + "\r\nContent-length: ".length();

    //read content length
    if ( (crlfpos = text.substring(pos).indexOf("\r\n")) == -1) {
        System.err.println("parseMessage: Error parsing line break after Content-length header");
        return null;
    }
    clen = Integer.parseInt(text.substring(pos, pos + crlfpos));
    pos += crlfpos + 2;

    if (text.substring(pos).toLowerCase().startsWith("var: ") == false) {
      System.err.println("parseMessage: Error parsing var header");
      return null;
    }

    //read content
    var = new String(text.substring(pos+5)); //+5: strlen("var: ")
    return var;
  }


  /**
   * Reads a text message from socket and returns the correspondent ConfMsg object. First, the
   * message type is read (3 characters). If it is an "AOK" or "AER" message, then (normally) there
   * are no more data to be read, so a ConfMsg object is returned with mtype set to ACTION_AOK or ACTION_AER and
   * mtext set to "AOK" or "AER" respectively. In other case, the content-length field is read and the remaining
   * data (whose length is specified by the content length) are read.
   * @param socket Socket
   * Socket where the data will be read from.
   * @return ConfMsg
   */
  public ConfMsg readMessage(Socket socket) {
    BufferedReader reader = null;
    String line = null;
    ConfMsg msg = null;
    StringBuffer textMessage = new StringBuffer();
    int clen = 0; //content length
    char content[];
    char strType[] = new char[3];
    try {
      reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
    }
    catch (IOException e) {
      System.err.println("readMessage: Error obtaining socket input stream");
      return null;
    }

    msg = new ConfMsg();
    try {
      //first read message type
      reader.read(strType, 0, 3);
      line = new String(strType);
      if (line == null) {
        System.err.println("readMessage: Error reading message type");
        return null;
      }

      if (line != null) {
        if (line.substring(0, 3).compareToIgnoreCase("ACK") == 0) {
          msg.mtype = ACTION_ACK;
        }
        else
        if (line.substring(0, 3).compareToIgnoreCase("ERR") == 0) {
          msg.mtype = ACTION_ERR;
        }
        else
        if (line.substring(0, 3).compareToIgnoreCase("AOK") == 0) {
          msg.mtype = ACTION_AUTH_OK;
        }
        else
        if (line.substring(0, 3).compareToIgnoreCase("AER") == 0) {
          msg.mtype = ACTION_AUTH_ERR;
        }
        else {
          //unknown message type
          System.err.println("readMessage: Unknown message type - " +
                             line.substring(0, 3));
          return null;
        }
      }

      if ( (msg.mtype == ACTION_AUTH_OK) || (msg.mtype == ACTION_AUTH_ERR) ) {
        msg.setText(line.substring(0, 3));
        return msg;
      }

      //read the crlf
      if (reader.readLine().equals("") == false) {
        //malformed message: mtype should be followed by a crlf
        return null;
      }

      textMessage.append(line + "\r\n");

      //then read content length
      line = reader.readLine();
      if (line == null) {
        System.err.println("readMessage: Error reading content length");
        return null;
      }
      if (line != null) {
        if (line.substring(0,
            "Content-length: ".length()).compareToIgnoreCase("Content-length: ") !=
            0) {
          System.err.println("readMessage: Error reading content length value");
          return null;
        }
        else {
          try {
            clen = Integer.parseInt(line.substring("Content-length: ".length()));
          }
          catch (NumberFormatException e) {
            System.err.println("Error parsing content length");
            return null;
          }
        }
      }
      textMessage.append(line + "\r\n");

      //finally, read (the remaining) content-length bytes
      content = new char[clen];
      reader.read(content, 0, clen);
    }
    catch (IOException e) {
      System.err.println("Error reading message");
      return null;
    }
    msg.setText(textMessage + new String(content));
    return msg;
  }

  /**
   * Sends an "AUT" message to the server and gets the server response. This message repsesents an authenrication request.
   * The user sends her passwd, the server checks the NVRAM "passwd" variable, and, if they match, sends an "AOK" message.
   * Otherwise, it sends an "AER" message.
   * @param passwd String
   * The Linksys WRT54GS router administrator password sent by the user.
   * @param socket Socket
   * The socket over which the message will be sent.
   * @return boolean
   * Returns true if authentication succeeded.
   */
  public boolean authenticateUser(String passwd, Socket socket) {
    boolean authenticated = false;
    String text = null;
    ConfMsg msg = null;

    //send auth message to server
    text = makeMessage(ACTION_AUTH, passwd, null);
    if (text == null) {
      return false;
    }

    try {
      socket.getOutputStream().write(text.getBytes());
    }
    catch (IOException ex) {
      System.err.println("authenticateUser: Error writing to socket");
      return false;
    }

    //read server response
    msg = readMessage(socket);
    if (msg == null) {
      return false;
    }
    if (msg.mtype != ACTION_AUTH_OK) {
      System.err.println("!AOK");
      return false;
    }
    else {
      return true;
    }
  }

  /**
   * Connect to the configuration server. If a connection can be established, the resulting
   * socket is returned. The server address used is this.serverAddr and the port is this.CONF_SERVER_PORT.
   * @return Socket
   */
  public Socket connectToServer() {
    Socket socket = null;
    try {
      socket = new Socket(serverAddr, CONF_SERVER_PORT);
    }
    catch (IOException e) {
      System.err.println("connectToServer: Could not connect to server " +
                         serverAddr.getHostName() + " at port " + CONF_SERVER_PORT);
      return null;
    }
    return socket;
  }

  public static void main(String args[]) throws IOException {
    InetAddress srv = null;
    ConfClient client = null;
    Socket socket = null;
    String text = null;
    ConfMsg msg = null;

    try {
      srv = InetAddress.getByName("192.168.1.1");
    }
    catch (java.net.UnknownHostException e) {
      System.err.println("Error getting server address");
      System.exit(1);
    }
    client = new ConfClient(srv);
    socket = client.connectToServer();
    socket.setKeepAlive(true);
    if (!client.authenticateUser("pantelis", socket)) {
      System.err.println("Could not authenticate user");
      System.exit(1);
    }
    text = client.makeMessage(ACTION_SET, "paparia", "malakia");
    if (text == null) {
      System.err.println("null text message");
      System.exit(1);
    }
    try {
      socket.getOutputStream().write(text.getBytes());
    }
    catch (IOException ex) {
      System.err.println("Error writing to socket");
      System.exit(1);
    }

    msg = client.readMessage(socket);
    if (msg == null) {
      System.err.println("Null message from server");
      System.exit(1);
    }
    else {
      System.err.println(msg.getText());
    }
    socket.close();
  }
}
